"""
This module contains various functions and classes used internally
by the PacketLogic API. They are not designed for public use, and
their availability and behaviour may change without notice.

@var __revision__: Revision identifier.
"""

__revision__ = '%K%'

import select
import socket
import sys
import time

from thisversion import PLDB


# Using a too large timeout to connect() will result in a
# "OverflowError: timeout period too long" on Windows.
MAX_TIMEOUT = 60


def negotiate_flicka_with_timeout (sock, io, timeout):
	deadline = time.time() + (timeout or sys.maxint)
	sock.setblocking(False)

	while True:
		now = time.time()
		if now > deadline:
			raise IOError("Operation timed out")

		wr = io.negotiate_want_read() and [sock] or []
		ww = io.negotiate_want_write() and [sock] or []

		rr, rw, _ = select.select(wr, ww, [], min(deadline-now, MAX_TIMEOUT))

		res = io.negotiate(bool(rr), bool(rw))
		if res:
			break


def verify_ip6(value):
	if value.find(":") == -1:
		return False

	parts = value.split("::")
	if len(parts) > 2:
		return False

	filled = (len(parts) == 2)

	parts = value.split(":")
	cnt = 0
	for p in parts:
		if p == '':
			continue
		try:
			v = int(p, 16)
		except ValueError:
			return False
		if not (0 <= v < 0xffff):
			return False
		cnt += 1

	if (not filled) and cnt != 8:
		return False
	if filled and cnt > 7:
		return False

	return True
		
def verify_ip(ip):
	"""
	Verifies that a string is a IP address. This is done
	using inet_aton and watching for errors. This means that
	incomplete addresses (i.e. 192.168.0) will be accepted.

	@type ip: C{str}
	@param ip: The string to be tested.

	@rtype: C{boolean}
	@returns: True if the string was a IP or else False
	"""
	try:
		socket.inet_aton(ip)
	except socket.error:
		return False

	if not len(ip.split(".")) == 4:
		return False
	return True


class _Callback:
	def __init__(self, pldb):
		self._pldb = pldb
		self.done = False
		self.res = None

	def setpldb(self, o):
		o._pldb = self._pldb
		return o

	def handleOK(self):
		return self.res

	def handleData(self):
		raise NotImplementedError

	def handler(self, msg):
		if msg.get_type() == PLDB.MSG_TYPE_OK:
			self.res = self.handleOK()
			self.done = True
			return True
		else:
			self.handleData(msg)
			return False

class _ListCB(_Callback):
	def __init__(self, pldb, handler, typ, nopldb=True):
		_Callback.__init__(self, pldb)
		self.list = None
		self.h = handler
		self.typ = typ
		self.nopldb = nopldb

	def handleOK(self):
		if not self.list:
			return []

		if self.nopldb:
			return map(self.setpldb, self.list.convert(self.typ))
		else:
			return self.list.convert(self.typ)

	def handleData(self, msg):
		self.list = self.h(msg, self.list)

class _SimpleListCB(_Callback):
	def __init__(self, pldb, handler, *optional_args):
		_Callback.__init__(self, pldb)
		self.list = []
		self.h = handler
		self.optional_args = optional_args

	def handleOK(self):
		return self.list

	def handleData(self, msg):
		self.list += self.h(msg, *self.optional_args)


class _AddCB(_Callback):
	def __init__(self, pldb, handler=None):
		_Callback.__init__(self, pldb)
		self.r = None
		self.h = handler

	def handleOK(self):
		return self.r
	def handleData(self, msg):
		if self.h:
			self.r = self.h(msg)
		else:
			self.r = True

def _do(pldb, cid, cb):
	pldb._addcidcb(cid, cb.handler)

	while not cb.done:
		pldb._process()

	return cb.res


####### Stuff for read-only classes and properties etc

# Readonly mutable objects (using the same errors as a tuple raises if you try to edit it).
# If you want to change them (in internal code) use e.g: list.append(myRoList, 7)
def _noset(*args, **kwargs): raise TypeError('object does not support item assignment')
def _nodel(*args, **kwargs): raise TypeError('object does not support item deletion')

class RoDict(dict):
	""" Read-only version of C{dict}."""
	__delitem__ = _nodel
	__setitem__ = _noset
	setdefault = _noset
	clear = _noset
	update = _noset
	pop = _noset
	popitem = _noset
	
class RoList(list):
	""" Read-only version of C{list}."""
	__delitem__ = _nodel
	__delslice__ = _nodel
	remove = _nodel
	__add__ = _noset
	__setitem__ = _noset
	__setslice__ = _noset
	append = _noset
	extend = _noset
	insert = _noset
	pop = _noset
	reverse = _noset
	sort = _noset
	#This will allow e.g: for item in obj.items: obj.remove_item(item)
	def __iter__(self):
		return self[:].__iter__()

def deprecatedalias(oldnam, newnam):
	def _getter(self):
		import warnings
		warnings.warn("Usage of %(oldnam)s is deprecated, instead use %(newnam)s" % {'oldnam': oldnam, 'newnam': newnam},
			      DeprecationWarning, stacklevel=2)
		
		return getattr(self, newnam)
	return property(_getter, doc="""Deprecated alias for %(newnam)s

	                          @deprecated: Usage of %(oldnam)s is deprecated, instead use %(newnam)s""" % {'oldnam': oldnam, 'newnam': newnam})

# Shortcuts for readonly-properties that makes class definitions less cluttery.
# To set a variable internally, use e.g self._prop.property = value

roConst = lambda value,docstr=None: property(lambda self, const=value: const, doc=docstr)
roConstList = lambda value,docstr=None: property(lambda self, const=RoList(value): const, doc=docstr)
roConstDict = lambda value,docstr=None: property(lambda self, const=RoDict(value): const, doc=docstr)
roCopyList = lambda key,docstr=None: property(lambda self: self._prop(key)[:], doc=docstr)
roCopyDict = lambda key,docstr=None: property(lambda self: self._prop(key).copy(), doc=docstr)

#roProp = lambda key,docstr=None: property(lambda self: self._prop(key), doc=docstr)
#roRoList = lambda key: property(lambda self: RoList(self._prop(key)))
#roRoDict = lambda key: property(lambda self: RoDict(self._prop(key)))

"""
Function that constructs a _Prop class with the named properties.

This class is only used internally.
	
Example:
	MyProp = plutil.Prop('one', 'two', three=list, four=plutil.RoDict)
	_prop = MyProp(one=1, four={'apa': 'korv'])
		
<names> is a list of properties that will be included.
<namevals> is a dict of properties to be included, with a callable constructor.
           The constructor typically is a type, that should accept zero or one argument.
		   
When a new instance is created, all property in <namevals> will be created, while
properties in <names> will only be created if specified in the init-call.
If the <namevals> items are specified, as 'four' in the example, that value will be used
as argument to the constructor, otherwise the constructor is called without argument.
The constructor is also run if you set the attribute with e.g _prop('three', [1, 7])
or _prop.update(three=[1, 7]).
"""

def Prop(*names, **namevals):
	"""
	Function that creates classed used internally for storing properties.
	"""
	class _Prop(object):
		__slots__ = tuple(names) + tuple(namevals.keys())
		_constructors = namevals
		
		def __init__(self, **_namevals):
			""" Any of the properties can be supplied as a keyword argument to be set.
			If the property has a constructor defined it will be run on the value.
			Properties with constructors defined, that is not supplied as argument, will
			be set to an empty value (i.e the constructor will be run without argument).
			"""
			[setattr(self, k, v()) for (k, v) in self._constructors.items() if k not in _namevals]
			self.update(**_namevals)
			
		def update(self, **_namevals):
			""" Will set the properties supplied as keyword arguments.
			If the property has a constructor defined it will be run on the value.
			"""
			[setattr(self, k, map(self._constructors.get(k), [v])[0]) for (k, v) in _namevals.items()]
				
		def __iter__(self):
			""" Return an iterator of all (prop, value)."""
			# Unfortunately python converts single item __slots__ to a string instead of a one item tuple.
			if isinstance(self.__slots__, tuple):
				return [(k, getattr(self, k)) for k in self.__slots__].__iter__()
			else:
				return [(self.__slots__, getattr(self, self.__slots__))]

		def __call__(self, key, *arg):
			""" Will return the named property. If the optional value is supplied
			it will set the property first, using the constructor if defined.
			"""
			if len(arg) == 1:
				#setattr(self, key, map(self._constructors.get(key), [arg[0]])[0])
				if key in self._constructors:
					setattr(self, key, self._constructors[key](arg[0]))
				else:
					setattr(self, key, arg[0])
			
			elif len(arg):
				raise TypeError, "%s expected at most 3 arguments, got %d" % (self.__name__, len(arg) + 2)
			
			return getattr(self, key)
				
	return _Prop

class BaseFlags(object):
	""" Subclasses of this is used internaly to represent flags."""
	__slots__ = ('_prop')
	_Prop_class = None
	
	def __init__(self, *args, **kwargs):
		""" See _set_flags for explanation of arguments."""
		if args:
			self._prop = self._Prop_class()
			self._set_flags(*args, **kwargs)
		else:
			self._prop = self._Prop_class(**kwargs)

	def items(self):
		""" Return a list of (prop, value) tuples, similar to dict.items()."""
		return list(self._prop)

	def __iter__(self):
		return [k for (k, v) in self._prop if v].__iter__()

	def __repr__(self):
		l = ["%s=%s" % (k, v) for (k, v) in self._prop if v]
		return "%s(%s)" % (self.__class__.__name__, (", ".join(l)) or 'None')

	def _set_flags(self, *reset, **flags):
		"""
		Set or reset one or more of the flags, named in kwargs.
		You may also reset all flags with the single argument None,
		Or clone an existing flags object.
		You may combine them, e.g (None, onlythisflag=True).
		"""
		if len(reset) > 1:
			raise TypeError("set_flags() takes at most 1 non-keyword argument. %d given." % len(reset))
		
		if len(reset):
			for flag in self:
				self._prop(flag, False)

			if isinstance(reset[0], (BaseFlags, list)):
				# Supply another flags class to clone, or a list of flags that should be set.
				for flag in reset[0]:
					self._prop(flag, True)

			elif reset[0] is not None:
				self._set_flags_other_type(reset[0])
		
		for flag, state in flags.items():
			self._prop(flag, state)
	
	def _set_flags_other_type(self, flags):
		""" Can be subclassed if we support something else as positional argument."""
		raise ValueError("First argument, if used, should be None or another flags class to clone.")
		
"""
Example:

import plutil

# Name all properties that should reside in the _prop object.
_Prop_apa = plutil.Prop('id', 'name', 'items', members=plutil.RoList)

# Use new style class (i.e subclass of 'object').
# Supply a __slot__ with '_prop' and optionally any additional read-write properties you will set in e.g __init__()

class Apa(object):
	__slots__ = ('_prop', '_pldb')
	def __init__(self, name):
		# Set self._prop to the class created before, and optionally preset some values in one of the ways shown here.
		self._prop = _Prop_apa(items = [])
		self._prop.name = name
		
		self._pldb = None

	def add_item(self, item):
		# Example of how you change a readonly property internally
		self._prop.items.append(item)

	def add_member(self, item):
		# Example of how you change a RoList property internally
		list.append(self._prop.members, item)

	def __repr__(self):
		# Example of various ways to read properties internally
		return '<Apa #%s: %s, %s>' % (self.id, self._prop('name'), self._prop.items)

	# Method for the create-on-call property. Preferably named underscore followed by the property name.
	def _more_items(self):
		return [item + item for item in self._prop.items]
	
	### Define all (readonly) properties.
	
	# Constants can't even be changed internally later, and appear only here.
	opts = plutil.roConstDict({0: 'eq', 1: 'ne'})
	
	# Example of a property that is created when we access it, and isn't just a value. (not listed in _Prop_apa class.)
	more_items = property(_more_items)
	
	# Other properties should be exactly the same that we named in the _Prop_apa class before.
	id = property(lambda self: self._prop.id)
	name = property(lambda self: self._prop.name)
	items = plutil.roCopyList('items')  # A copy ot the internal list, you can edit it, but the internal list is unaffected.
	members = property(lambda self: self._prop.members)


Example of a simple flags class:

_Prop_ShapingObjectFlags = plutil.Prop(brown=bool, host_fairness=bool)
class ShapingObjectFlags(plutil.BaseFlags):
	__slots__ = ('_prop')
	_Prop_class = _Prop_ShapingObjectFlags
		
	brown = property(lambda self: self._prop.brown)
	host_fairness = property(lambda self: self._prop.host_fairness)

Sometimes you would like to e.g add an __int__() method and support to
set the flags using the API integer. that could add these two methods:

	def _set_flags_other_type(self, flags):
		if isinstance(flags, (int, long)):
			self._prop.update(brown=flags & 0x01, host_fairness=flags & 0x02)
		else:
			plutil.BaseFlags._set_flags_other_type(self, flags)

	def __int__(self):
		return self._prop.brown * 0x01 + self._prop.host_fairness * 0x02
"""

class AttributesDict(dict):
	"""
	Dictionary used for pldbobject.attributes dictionary,
	changes 
	"""
	def __init__(self, l, sethook, removehook):
		dict.__init__(self, l)
		self.sethook = sethook
		self.removehook = removehook
	def __setitem__(self, k, v):
		if not isinstance(k, basestring):
			raise TypeError("Key must be string")
		if not isinstance(v, basestring):
			raise TypeError("Value must be string")
		self.sethook(k, v)
		dict.__setitem__(self, k, v)
	def __delitem__(self, k):
		self.removehook(k)
		dict.__delitem__(self, k)
	def not_implemented(self, *k, **kv):
		raise NotImplementedError("")
	clear = not_implemented
	copy = not_implemented
	pop = not_implemented
	popitem = not_implemented
	update = not_implemented
