"""
This module contains functionality for managing channels on the
PacketLogic system.

Do NOT import this module directly. The L{PLConnection
<packetlogic2.PLConnection>} object should be used, e.g:

    >>> import packetlogic2
    >>> pl = packetlogic2.connect("192.168.1.25", "admin", "pldemo00")
    >>> ch = pl.Channels()

@var __revision__: Revision identifier.
"""

__revision__ = '%K%'

from resource import Resource
from thisversion import PLDB

import plutil
from packetlogic2.exceptions import PLDBUnsupportedInFirmware

_Prop_Channel = plutil.Prop('id', 'description', 'internal', 'internal_media', 'external', 'external_media', active=bool)

class Channel(object):
	""" Information about a Channel.
	
	Returned in a list by L{Channels.channel_list}.
	
	@type id: C{int}
	@type active: C{bool}
	@type label: C{str}
	@type location: C{str}
	@type description: C{str}
	@type internal: C{str}
	@type internal_media: C{int}
	@type external: C{str}
	@type external_media: C{int}
	
	@cvar __slots__: For internal use only.
	"""
	__slots__ = ('_prop', '_pldb')
	
	def __init__(self, _pldb=None, **props):
		"""
		Should not be instantiated manually, use L{channel_list <Channels.channel_list>}.
		"""
		self._pldb = _pldb
		self._prop = _Prop_Channel(**props)

	def __repr__(self):
		return "<%s #%d '%s'>" % (self.__class__.__name__, self._prop.id, self._prop.description)

	def set_active(self, active):
		""" Set or reset the active state on this Channel.

		@type active: C{bool}
		@param active: The active state to set on this Channel.

		"""
		active = bool(active)
		plutil._do(self._pldb, self._pldb.session.channel_set_active(self._prop.id, active), plutil._AddCB(self._pldb))
		self._prop.active = active

	def update(self, internal_media=None, external_media=None, role=None, label=None):
		""" Update properties on this Channel.
		Omitting an argument or setting it to C{None} will leave that property unchanged.

		@type internal_media: C{int}
		@param internal_media: One of the CHANNEL_MEDIA_ constants.

		@type external_media: C{int}
		@param external_media: One of the CHANNEL_MEDIA_ constants.

		@type role: C{int}
		@param role: One of the ROLE_ constants.

		@type label: C{str}
		@param label: Label for this channel.

		"""
		if internal_media is None:
			internal_media = self._prop.internal_media

		if external_media is None:
			external_media = self._prop.external_media

                if role is not None:
                        raise PLDBUnsupportedInFirmware("Usage of 'role' is not supported in this firmware version")
                if label is not None:
                        raise PLDBUnsupportedInFirmware("Usage of 'label' is not supported in this firmware version")

		plutil._do(self._pldb,
                           self._pldb.session.channel_update(self, internal_media, external_media), plutil._AddCB(self._pldb))

		self._prop.internal_media = internal_media
		self._prop.external_media = external_media

        def _raise_label(self):
                raise PLDBUnsupportedInFirmware("Usage of 'label' is not supported in this firmware version")
        def _raise_caps(self):
                raise PLDBUnsupportedInFirmware("Usage of 'caps' is not supported in this firmware version")
        def _raise_role(self):
                raise PLDBUnsupportedInFirmware("Usage of 'role' is not supported in this firmware version")

	id = property(lambda self: self._prop.id, doc="ID Number of the channel.")
	label = property(_raise_label,
                         doc="""The channel's label.

                             Label used for referencing this channel
                             from filtering rules.

                             @attention: label requires firmware 14.1 or newer.

                             """)
	location = property(lambda self: self._prop.description,
                            doc="""The channel's location.

                                A string describing the channels location.
                                e.g: "Channel 1/2"
                                """)
	description = property(lambda self: self._prop.description,
                               doc="Deprecated alias for 'location'.")
	active = property(lambda self: self._prop.active, doc="Active state.")
	internal = property(lambda self: self._prop.internal, doc="System name of the internal interface.")
	internal_media = property(lambda self: self._prop.internal_media, doc="Media type on Internal interface. One of the MEDIA_ constants.")
	external = property(lambda self: self._prop.external, doc="System name of the external interface.")
	external_media = property(lambda self: self._prop.external_media, doc="Media type on External interface. One of the MEDIA_ constants.")
	caps = property(_raise_caps, doc="Capabilities bitmask. One or more of the CAPS_ constants.")
	role = property(_raise_role, doc="Channel role. One of the ROLE_ constants.")

class Channels(Resource):
	""" This is the resource for Channel inquiry and manipulation.
	Changes needs to be commited to take effect.

	@cvar CHANNEL_MEDIA_AUTO: Media type is Auto Negotiated
	@cvar CHANNEL_MEDIA_10HD: Media type is 10 Mbps Half Duplex
	@cvar CHANNEL_MEDIA_10FD: Media type is 10 Mbps Full Duplex
	@cvar CHANNEL_MEDIA_100HD: Media type is 100 Mbps Half Duplex
	@cvar CHANNEL_MEDIA_100FD: Media type is 100 Mbps Full Duplex
	@cvar CHANNEL_MEDIA_1000FD: Media type is 1 Gbps Full Duplex
	@cvar CHANNEL_MEDIA_10000FD: Media type is 10 Gbps Full Duplex

	@group Channels: channel_list
	
	"""
	
        CHANNEL_MEDIA_AUTO = PLDB.CHANNEL_MEDIA_AUTO
        CHANNEL_MEDIA_10HD = PLDB.CHANNEL_MEDIA_10HD
        CHANNEL_MEDIA_10FD = PLDB.CHANNEL_MEDIA_10FD
        CHANNEL_MEDIA_100HD = PLDB.CHANNEL_MEDIA_100HD
        CHANNEL_MEDIA_100FD = PLDB.CHANNEL_MEDIA_100FD
        CHANNEL_MEDIA_1000FD = PLDB.CHANNEL_MEDIA_1000FD
        CHANNEL_MEDIA_10000FD = PLDB.CHANNEL_MEDIA_10000FD
	
	def _resource(self):
		return PLDB.RESOURCE_CHANNELS

	def channel_list(self):
		"""
		List all Channels (Traffic Interface Pairs) in the PacketLogic system.

			>>> ch.channel_list()
			[<Channel #299 'Slot 4/GE.5 (Reserved)'>,
			 <Channel #216 'Slot 4/GE.4'>,
			 <Channel #212 'Slot 4/GE.3'>,
			 <Channel #208 'Slot 4/GE.2'>,
			 <Channel #204 'Slot 4/GE.1'>,
			 <Channel #200 'Slot 4/10GE'>,
			 <Channel #199 'Slot 3/GE.5 (Monitor/FlowSync)'>,
			 <Channel #116 'Slot 3/GE.4'>,
			 <Channel #112 'Slot 3/GE.3'>,
			 <Channel #108 'Slot 3/GE.2'>,
			 <Channel #104 'Slot 3/GE.1'>,
			 <Channel #100 'Slot 3/10GE'>]

		
		@rtype: C{list} of L{Channel}
		@returns: List containing a L{Channel} object for each channel available on the system.
		"""

		return plutil._do(self, self.session.channel_list(),
				  plutil._ListCB(self, self.session.channel_list_handle, Channel))
