"""
This module contains functionality for accessing the "File server" on
the PacketLogic system.

Do NOT import this module directly. The L{PLConnection
<packetlogic2.PLConnection>} object should be used, e.g:

    >>> import packetlogic2
    >>> pl = packetlogic2.connect("192.168.1.25", "admin", "pldemo00")
    >>> fs = pl.Fileserv()

@var __revision__: Revision identifier.
"""

__revision__ = '%K%'

import tempfile

import resource
from thisversion import PLDB

import plutil

def _fileservtype(name):
    def _getter(self):
        import warnings
        warnings.warn("Deprecated Fileserv area ID constant used. Use string instead (i.e %s)" % repr(name),
                      DeprecationWarning, stacklevel=1)
        return self._ensure_numeric_ftype(name)

    return property(_getter, doc="Deprecated ID number used to access files in the '%s' area. Use string instead." % name)

class Fileserv (resource.Resource):
    """
    This resource gives access to the "Fileserv" on the PacketLogic
    system, allowing up- and download of files.

    @group Fileserv: types,list,download,upload,remove
    """

    HOSTTRIGGER     = _fileservtype("Host trigger files")
    FWTRIGGER       = _fileservtype('Firewall trigger files')
    DHCP_SNOOPER    = _fileservtype('DHCP snooper files')
    RADIUS_SNOOPER  = _fileservtype('RADIUS snooper files')
    SIP_SNOOPER     = _fileservtype('SIP snooper files')
    SNMP_MIB        = _fileservtype('SNMP')
    LICENSE_UPLOAD  = _fileservtype('License upload')
    DOCUMENTATION   = _fileservtype('Documentation')
    PCAP_1          = _fileservtype('PCAP writer files')
    PCAP_2          = _fileservtype('PCAP-2 writer files')
    FIRMWARE_UPLOAD = _fileservtype('Upgrade files')
    ATTRIBUTES      = _fileservtype('Attribute descriptions')
    BOOKMARKS       = _fileservtype('Bookmarks')

    def _resource (self):
        return PLDB.RESOURCE_FILESERV

    # PYTHONIC API
    def types (self):
        """
        List available file types (directories) available on the system.

        >>> fs.types()
        [{'flags': 3L, 'id': 0L, 'name': 'Host Trigger Files'},
         {'flags': 3L, 'id': 1L, 'name': 'Firewall Trigger Files'},
         {'flags': 3L, 'id': 2L, 'name': 'DHCP Snooping files'},
         {'flags': 3L, 'id': 3L, 'name': 'RADIUS Snooping files'},
         {'flags': 3L, 'id': 4L, 'name': 'SIP Snooper files'},
         {'flags': 1L, 'id': 5L, 'name': 'SNMP'},
         {'flags': 3L, 'id': 6L, 'name': 'License upload'},
         {'flags': 1L, 'id': 7L, 'name': 'Documentation'},
         {'flags': 1L, 'id': 8L, 'name': 'PCAP Writer files'},
         {'flags': 1L, 'id': 9L, 'name': 'PCAP-2 Writer files'},
         {'flags': 3L, 'id': 10L, 'name': 'Upgrade files'},
         {'flags': 3L, 'id': 11L, 'name': 'Attribute descriptions'}]

         @rtype: C{list} of C{dict}
         @returns: List of the different file types available.
                   Each type is represented as a C{dict} with
                   the keys C{'name'}, C{'id'} and C{'flags'}

        """
        return self._fileserv_list_types()

    def list (self, ftype):
        """
        List all files of specified type.

            >>> fs.list(5)
            [{'r_path': 'PACKETLOGIC-MIB'}]

            >>> fs.list(fs.SNMP_MIB)
            [{'r_path': 'PACKETLOGIC-MIB'}]

            >>> fs.list('SNMP')
            [{'r_path': 'PACKETLOGIC-MIB'}]

        @rtype: C{list} of C{dict}
        @returns: List of files. Each file is represented as a C{dict} with the key C{'r_path'}
        """
        ftype = self._ensure_numeric_ftype(ftype)
        return self._fileserv_list_files(ftype)

    def download (self, ftype, fname):
        """
        Download specified file.

            >>> mib = fs.download('SNMP', 'PACKETLOGIC-MIB')
            >>> len(mib)
            191948
            >>> mib[:84]
            '-- THIS FILE IS AUTO-GENERATED - DO NOT EDIT!\\n\\nPACKETLOGIC-MIB DEFINITIONS ::= BEGIN'

        @type ftype: C{int} or C{str}
        @param ftype: Type of file, the name of the area as a string,
                      the 'id' provided by the L{types() method
                      <Fileserv.types>} or one of the constants on the
                      Fileserv object.

        @type fname: C{str}
        @param fname: Name of file.

        @rtype: C{str}
        @returns: The contents of the requested file.
        """
        ftype = self._ensure_numeric_ftype(ftype)

        ofile = tempfile.TemporaryFile()
        size = self._fileserv_download(ftype, fname, ofile)
        ofile.flush()
        ofile.seek(0)
        data = ofile.read()
        assert len(data) == size
        
        return data

    def upload (self, ftype, fname, data):
        """
        Upload a file to fileserv.

            >>> licensedata=file('001122334455').read()
            >>> fs.upload('License upload', '001122334455.lic', licensedata)

        @type ftype: C{int}
        @param ftype: Type of file, the name of the area as a string,
                      the 'id' provided by the L{types() method
                      <Fileserv.types>} or one of the constants on the
                      Fileserv object.

        @type fname: C{str}
        @param fname: Name of file.

        @type data: C{str}
        @param data: The contents of the file.

        @rtype: C{str}
        @returns: The contents of the requested file.
        """
        ftype = self._ensure_numeric_ftype(ftype)

        ifile = tempfile.TemporaryFile()
        ifile.write(data)
        ifile.flush()
        ifile.seek(0)
        self._fileserv_upload(ftype, fname, ifile)

    def remove (self, ftype, fname):
        """
        Remove specified file from fileserv.

        @type ftype: C{int}
        @param ftype: Type of file, the name of the area as a string,
                      the 'id' provided by the L{types() method
                      <Fileserv.types>} or one of the constants on the
                      Fileserv object.

        @type fname: C{str}
        @param fname: Name of file.
        """
        ftype = self._ensure_numeric_ftype(ftype)
        self._fileserv_remove(ftype, fname)

    # HELPERS
    def _ensure_numeric_ftype(self, ftype):
        if isinstance(ftype, basestring):
            for t in self.types():
                if t['name'].lower() == ftype.lower():
                    return t['id']
            raise KeyError("Could not find requested file type ('%s')" % ftype)
        return ftype

    # UNPYTHONIC API
    def _fileserv_list_types (self):
        cid = self.session.fileserv_types()
        cb  = plutil._ListCB(self, self.session.fileserv_types_handle,
                             PLDB.PLFileservTypeList, False)
        return plutil._do(self, cid, cb)

    def _fileserv_list_files (self, *args):
        cid = self.session.fileserv_list(*args)
        cb  = plutil._ListCB(self, self.session.fileserv_list_handle,
                             PLDB.PLFileservList, False)
        return plutil._do(self, cid, cb)

    def _fileserv_download (self, ftype, fname, ofile):
        assert hasattr(ofile, 'write')
        
        def callback (msg):
            if msg.get_type() == PLDB.MSG_TYPE_DATA:
                ok, callback.size = self.session.fileserv_download_handle(
                    msg, ofile, callback.size)
                if not ok:
                    raise resource.PLDBError (ok, callback.size)
            callback.done = msg.get_type() == PLDB.MSG_TYPE_OK
            return callback.done

        callback.size = 0
        callback.done = False
        
        cid = self.session.fileserv_download(ftype, fname)
        self._addcidcb(cid, callback)

        while not callback.done:
            self._process()

        return callback.size

    def _fileserv_remove (self, ftype, fname):
        def callback (msg):
            callback.done = True
            if msg.get_type() == PLDB.MSG_TYPE_OK:
                self.ok = True
            else:
                self.ok = False
        callback.done = False

        cid = self.session.fileserve_remove(ftype, fname)
        self._addcidcb(cid, callback)
        while not callback.done:
            self._process()
            
    def _fileserv_upload (self, ftype, fname, ifile):
        assert hasattr(ifile, 'read')
        cid, fil = self.session.fileserve_upload(ftype, fname, ifile)

        def callback (msg):
            cid = self.session.fileserve_upload_handle(msg, fil)
            if cid:
                self._addcidcb(cid, callback)
            else:
                callback.done = True
        callback.done = False

        self._addcidcb(cid, callback)
        while not callback.done:
            self._process()


