"""
This module contains the implementation of the different rule types
(L{FwRule}, L{ShapingRule} L{StatisticsRule}).

There is no need to import this module directly. Use the methods of
the L{Ruleset object <ruleset.Ruleset>} to create and manipulate
rules.

@var __revision__: Revision identifier.
"""

__revision__ = '%K%'

from thisversion import PLDB
import plutil

from packetlogic2.exceptions import PLDBUnsupportedInFirmware

NONEXISTANT_CONDITION_TYPE = object()
NONEXISTANT_FWRULE_ACTION = object()

def _fwlist(pldb):
	return plutil._do(pldb, pldb.session.fwrule_list(), plutil._ListCB(pldb, pldb.session.fwrule_list_handle, lambda **a: FwRule(pldb, **a)))

def _shapinglist(pldb):
	return plutil._do(pldb, pldb.session.shapingrule_list(), plutil._ListCB(pldb, pldb.session.shapingrule_list_handle, lambda **a: ShapingRule(pldb, **a)))

def _statisticslist(pldb):
	return plutil._do(pldb, pldb.session.statisticsrule_list(), plutil._ListCB(pldb, pldb.session.statisticsrule_list_handle, lambda **a: StatisticsRule(pldb, **a)))

def _monitoriface_list(pldb):
	return plutil._do(pldb, pldb.session.monitorinterface_list(), plutil._ListCB(pldb, pldb.session.monitoriface_list_handle, MonitorIface, False))

def _divertiface_list(pldb):
	return plutil._do(pldb, pldb.session.divertinterface_list(), plutil._ListCB(pldb, pldb.session.divertiface_list_handle, DivertIface, False))

def _fwadd(_pldb, name, action, quick, log_level, rewriteobject=0, monitor_iface=0, monitor_label="", divert_label="", inject_data="", trigger=""):
	kwargs = locals().copy()
	cid = _pldb.session.fwrule_add(name, action, quick, log_level, rewriteobject, monitor_iface, monitor_label, divert_label, inject_data, trigger)
	(id, position, tstamp) = plutil._do(_pldb, cid, plutil._AddCB(_pldb, _pldb.session.fwrule_add_handle))
	res = FwRule(
		id = id,
		position = position,
		orderby = 0,
		conditions = {},
		active = True,
		creator = _pldb.user,
		creation_date = tstamp,
		modified_by = _pldb.user,
		modification_date = tstamp,
		**kwargs
		)

	return res


def _fwupdate(_pldb, i, action, quick, log_level, rewriteobject, monitor_iface, monitor_label, divert_label, inject_data, trigger):
	if log_level not in [0,1,2]:
		raise ValueError("Unknown log level")
	cid = _pldb.session.fwrule_update(i, action, quick, log_level, rewriteobject, monitor_iface, monitor_label, divert_label, inject_data, trigger)
	res = plutil._do(_pldb, cid, plutil._AddCB(_pldb, _pldb.session.fwrule_update_handle))
	return res
	
def _statisticsupdate(_pldb, i, connlog):
	cid = _pldb.session.statisticsrule_update(i, connlog)
	res = plutil._do(_pldb, cid, plutil._AddCB(_pldb, _pldb.session.statisticsrule_update_handle))
	return res
	
	


	
def _shapingadd(_pldb, name, priority=5):
	kwargs = locals().copy()
	cid = _pldb.session.shapingrule_add(name, priority)
	(id, tstamp) = plutil._do(_pldb, cid, plutil._AddCB(_pldb, _pldb.session.shapingrule_add_handle))
	res = ShapingRule(id = id,
			  active = True,
			  creator = _pldb.user,
			  creation_date = tstamp,
			  modified_by = _pldb.user,
			  modification_date = tstamp,
			  **kwargs
			  )
	return res

def _statisticsadd(_pldb, name):
	kwargs = locals().copy()
	cid = _pldb.session.statisticsrule_add(name, 0)
	(id, tstamp) = plutil._do(_pldb, cid, plutil._AddCB(_pldb, _pldb.session.statisticsrule_add_handle))
	res = StatisticsRule(id = id,
			     active = True,
			     creator = _pldb.user,
			     creation_date = tstamp,
			     modified_by = _pldb.user,
			     modification_date = tstamp,
			     **kwargs
			     )
	return res

def _fwremove(pldb, obj):
	cid = pldb.session.fwrule_remove(obj._prop.id)
	return plutil._do(pldb, cid, plutil._AddCB(pldb))

def _shapingremove(pldb, obj):
	cid = pldb.session.shapingrule_remove(obj._prop.id)
	return plutil._do(pldb, cid, plutil._AddCB(pldb))

def _statisticsremove(pldb, obj):
	cid = pldb.session.statisticsrule_remove(obj._prop.id)
	return plutil._do(pldb, cid, plutil._AddCB(pldb))

def _fwrename(pldb, obj, newname):
	cid = pldb.session.fwrule_rename(obj._prop.id, newname)
	return plutil._do(pldb, cid, plutil._AddCB(pldb, pldb.session.fwrule_rename_handle))

def _shapingrename(pldb, obj, newname):
	cid = pldb.session.shapingrule_rename(obj._prop.id, newname)
	return plutil._do(pldb, cid, plutil._AddCB(pldb, pldb.session.shapingrule_rename_handle))

def _statisticsrename(pldb, obj, newname):
	cid = pldb.session.statisticsrule_rename(obj._prop.id, newname)
	return plutil._do(pldb, cid, plutil._AddCB(pldb, pldb.session.statisticsrule_rename_handle))

def _fwswap(pldb, obj1, obj2):
	cid = pldb.session.fwrule_position_swap(obj1._prop.id, obj2._prop.id)
	return plutil._do(pldb, cid, plutil._AddCB(pldb, pldb.session.fwrule_position_swap_handle))

_Prop_ShapingRuleFlags = plutil.Prop(exclusive=bool, stats=bool, stats_netobjects=bool, stats_services=bool,
				     stats_local_vhosts=bool, stats_remote_vhosts=bool)
class ShapingRuleFlags(plutil.BaseFlags):
	""" Used internaly to represent the flags on a L{ShapingRule}.
	
	@type exclusive: C{bool}
	@type stats: C{bool}
	@type stats_netobjects: C{bool}
	@type stats_services: C{bool}
	@type stats_local_vhosts: C{bool}
	@type stats_remote_vhosts: C{bool}
	"""
	__slots__ = ('_prop')
	
	def __init__(self, **kwargs):
		self._prop = _Prop_ShapingRuleFlags(**kwargs)
		
	exclusive = property(lambda self: self._prop.exclusive, doc="This rule is used exclusively")
	stats = property(lambda self: self._prop.stats, doc="Enable collection of statistic on this rule")
	stats_netobjects = property(lambda self: self._prop.stats_netobjects, doc="Collect statistics on NetObjects")
	stats_services = property(lambda self: self._prop.stats_services, doc="Collect statistics on Services")
	stats_local_vhosts = property(lambda self: self._prop.stats_local_vhosts, doc="Collect statistics on Local VHosts")
	stats_remote_vhosts = property(lambda self: self._prop.stats_remote_vhosts, doc="Collect statistics on Remote VHosts")

_Prop_condition = plutil.Prop('op', 'type', 'objects')
_Prop_shapingrule = plutil.Prop('id', 'name', 'priority', 'active', 
				'creator', 'creation_date', 'modified_by', 'modification_date',
				conditions = plutil.RoDict, objects = plutil.RoList,
				flags = ShapingRuleFlags)
_Prop_fwrule = plutil.Prop('id', 'name', 'action', 'quick', 'log_level', 'rewriteobject',
			   'monitor_iface', 'position', 'orderby', 'trigger', 'active', 
			   'creator', 'creation_date', 'modified_by', 'modification_date', 'inject_data',
			   'divert_iface', 'monitor_label', 'divert_label', conditions = plutil.RoDict)


_PropatisticsRuleFlags = plutil.Prop(connlog=bool)
class StatisticsRuleFlags(plutil.BaseFlags):
	""" Used internaly to represent the flags on a L{StatisticsRule}.
	
	@type connlog: C{bool}
	"""
	__slots__ = ('_prop')
	
	def __init__(self, **kwargs):
		self._prop = _PropatisticsRuleFlags(**kwargs)
		
	connlog = property(lambda self: self._prop.connlog, doc="Connlog enabled on this rule")

_Prop_statisticsrule = plutil.Prop('id', 'name', 'active',
				   'creator', 'creation_date', 'modified_by', 'modification_date',
				   conditions = plutil.RoDict, objects = plutil.RoList, flags=StatisticsRuleFlags)


class PLRule(object):
	
	"""
	@note: This class is only used as a subclass and never on it's own.

	@type id: C{int}
	@type name: C{str}
	@type active: C{boolean}
	@type is_global: C{boolean}
	@type creator: C{str}
	@type creation_date: C{str}
	@type modified_by: C{str}
	@type modification_date: C{str}
	@type systemid: C{str}
	@type conditions: read-only C{dict} with C{int} keys and L{Contition <PLRule.Condition>} items.

	@type machineid: C{str}

	@cvar Condition: Condition implementation
	
	"""
	__slots__ = ()

	def cond_add(self, condition, operator, value):
		"""
		Adds a Condition to this rule

		@type condition: C{int}
		@param condition: One of the following condition types

			CONDITION_NETOBJECT_CLIENT
			CONDITION_NETOBJECT_SERVER
			CONDITION_NETOBJECT_HOST
			CONDITION_NETOBJECT_LOCAL
			CONDITION_PORTOBJECT_CLIENT
			CONDITION_PORTOBJECT_SERVER
			CONDITION_SERVICEOBJECT
			CONDITION_SERVICEOBJECT_BASE
			CONDITION_SYSTEMOBJECT
			CONDITION_TIMEOBJECT
			CONDITION_VLANOBJECT
			CONDITION_VLANOBJECT_IN
			CONDITION_VLANOBJECT_OUT
			CONDITION_PROTOCOLOBJECT
			CONDITION_TIMEOBJECT
			CONDITION_BGPOBJECT
			CONDITION_PROPERTYOBJECT
			CONDITION_FLAGOBJECT
			CONDITION_MPLSOBJECT
			CONDITION_MPLSOBJECT_IN
			CONDITION_MPLSOBJECT_OUT
			CONDITION_DSCPOBJECT_IN
			CONDITION_DSCPOBJECT_OUT

		@type operator: C{int}
		@param operator: One of CONDITION_OP_EQ (equals) or CONDITION_OP_NE (not equals)

		@type value: C{list} of C{int}
		@param value: A list of id's from the type chosen in the condition argument.

		"""
		raise NotImplementedError
	
	def cond_set(self, condition, operator, objects):
		"""
		Updates a Condition in this rule

		@type condition: C{int}
		@param condition: The condition type to modify. (CONDITION_*)

		@type operator: C{int}
		@param operator: One of CONDITION_OP_EQ (equals) or CONDITION_OP_NE (not equals)

		@type objects: C{list} of C{int}
		@param objects: A list of id's from the type chosen in the condition argument.

		"""
		if not condition in self._prop.conditions:
			raise ValueError("Condition does not exist")
		
		cb, cid = self._cond_set(self._prop.id, condition, operator, objects)
		tstamp = plutil._do(self._pldb, cid, plutil._AddCB(self._pldb, cb))
		self._prop.conditions[condition]._prop.op = operator
		self._prop.conditions[condition]._prop.objects = objects
		self._prop.modification_date = tstamp
		self._prop.modified_by = self._pldb.user
		
	def _cond_set(self, *args):
		raise NotImplementedError

	def cond_object_add(self, condition, objectid):
		"""
		Adds an object to a certain condition type

		@type condition: C{int}
		@param condition: The condition type to modify. (CONDITION_*)

		@type objectid: C{int}
		@param objectid: The id property of the suitable L{plobject.PLObject} implementation.

		"""
		if condition is NONEXISTANT_CONDITION_TYPE:
			raise PLDBUnsupportedInFirmware("Specified condition type is not supported in this firmware version")

		if not condition in self._prop.conditions:
			raise ValueError("Condition does not exist")

		cb, cid = self._cond_object_add(self._prop.id, condition, objectid)
		tstamp = plutil._do(self._pldb, cid, plutil._AddCB(self._pldb, cb))
		self._prop.conditions[condition]._prop.objects.append(objectid)
		self._prop.modification_date = tstamp
		self._prop.modified_by = self._pldb.user

	def _cond_object_add(self, *args):
		raise NotImplementedError
	
	def cond_object_remove(self, condition, objectid):
		"""
		Removes an object from a certain condition type

		@type condition: C{int}
		@param condition: The condition type to modify. (CONDITION_*)

		@type objectid: C{int}
		@param objectid: The id property of the suitable L{plobject.PLObject} implementation.

		"""
		if not condition in self._prop.conditions:
			raise ValueError("Condition does not exist")

		if not objectid in self._prop.conditions[condition]._prop.objects:
			raise ValueError("Condition object does not exist")

		cb, cid = self._cond_object_remove(self._prop.id, condition, objectid)
		tstamp = plutil._do(self._pldb, cid, plutil._AddCB(self._pldb, cb))
		self._prop.conditions[condition]._prop.objects.remove(objectid)
		self._prop.modification_date = tstamp
		self._prop.modified_by = self._pldb.user

	def _cond_object_remove(self, *args):
		raise NotImplementedError

	def cond_remove(self, condition):
		"""
		Removes a Condition from this rule

		@type condition: C{int}
		@param condition: Remove the condition type. (CONDITION_*)

		"""
		raise NotImplementedError

	def enable(self):
		"""
		Enables this rule

		"""
		raise NotImplementedError

	def disable(self):
		"""
		Disables this rule

		"""
		raise NotImplementedError

	def rename(self, newname):
		"""
		Rename the current rule.

		I{destroys netobject cache}

		@type newname: C{str}
		@param newname: The name that the rule should have after rename process

		"""
		raise NotImplementedError

	def _do_add(self, condition, operator, value, func, handlefunc, invalidator):
		"""
		Adds a Condition to this rule
		"""
		if not isinstance(value, (list, tuple)):
			raise TypeError("Argument <value> must be a list or tuple of object id:s.")

		if condition is NONEXISTANT_CONDITION_TYPE:
			raise PLDBUnsupportedInFirmware("Specified condition type is not supported in this firmware version")

		
		class _CondSetCb(plutil._Callback):
			def __init__(self, pldb):
				plutil._Callback.__init__(self, pldb)
				self.r = None
			def handleOK(self):
				return self.r
			def handleData(self, msg):
				self.r = handlefunc(msg)

		cid = func(self._prop.id, condition, operator, value)
		ret = plutil._do(self._pldb, cid, _CondSetCb(self._pldb))
		dict.__setitem__(self._prop.conditions, condition, self._create_cond(condition, operator, value))
		invalidator()
		return ret

	def _do_remove(self, condition, func, handlefunc, invalidator):

		if isinstance(condition, self.Condition):
			condition = condition._prop.type

		class _CondRemoveCb(plutil._Callback):
			def handleData(self, msg):
				self.res = handlefunc(msg)

		cid = func(self._prop.id, condition)
		ret = plutil._do(self._pldb, cid, _CondRemoveCb(self._pldb))
		dict.__delitem__(self._prop.conditions, condition)
		invalidator()
		return ret

	def _set_active(self, dofunc, handlefunc, arg):
		cid = dofunc(self._prop.id, arg)
		tstamp = plutil._do(self._pldb, cid, plutil._AddCB(self._pldb, handlefunc))
		self._prop.active = bool(arg)
		self._prop.modification_date = tstamp
		self._prop.modified_by = self._pldb.user


	def __repr__(self):
		return "<PacketLogic Rule '%s'>" % (self._prop.name)

	def _attr_set(self, key, value):
		ci = self._pldb.session.rule_attr_set(self._ruletype, self._prop.id, key, value)
		c = plutil._AddCB(self._pldb)
		plutil._do(self._pldb, ci, c)

	def _attr_remove(self, key):

		ci = self._pldb.session.rule_attr_remove(self._ruletype, self._prop.id, key)
		c = plutil._AddCB(self._pldb)
		plutil._do(self._pldb, ci, c)

	def _attr_list(self):
		ci = self._pldb.session.rule_attr_list(self._ruletype, self._prop.id)
		c = plutil._SimpleListCB(self._pldb, self._pldb.session.rule_attr_list_handle)
		c.list = []
		res = dict(plutil._do(self._pldb, ci, c))
		return plutil.AttributesDict(res, self._attr_set, self._attr_remove)
		

	def _create_cond(self, cond_type, op, objects):
		return self.Condition(
			pldb = self._pldb,
			type = cond_type,
			op = op,
			objects = objects,
			)

	def _machineid(self):
		""" Use of property 'machineid' is deprecated, use 'systemid' instead."""
		import warnings
		warnings.warn("Use of property 'machineid' is deprecated, use 'systemid' instead.",
			      DeprecationWarning, stacklevel=2)
		return self._pldb.systemid
	
	machineid = property(_machineid)
	
	systemid = plutil._sysidprop("System ID of the PacketLogic this rule belongs to")
	id = property(lambda self: self._prop.id, doc="id of rule")
	name = property(lambda self: self._prop.name, doc="Name of the rule. Changed with L{rename() <PLRule.rename>}")
	conditions = property(lambda self: self._prop.conditions, doc="Dict with condition type as key, and corresponding Condition as value.\n"
			      "Changed with L{cond_add() <PLRule.cond_add>}, L{cond_set() <PLRule.cond_set>}, L{cond_remove() <PLRule.cond_remove>},\n"
			      "L{cond_object_add() <PLRule.cond_object_add>} and L{cond_object_remove() <PLRule.cond_object_remove>}.")
	active = property(lambda self: self._prop.active, doc="Is the rule active.\n"
			  "Changed with L{enable() <PLRule.enable>} and L{disable() <PLRule.disable>}")
	is_global = property(lambda self: False, doc="True if this rule is on a remote system.")
	creator = property(lambda self: self._prop.creator, doc="Username that created this rule")
	creation_date = property(lambda self: self._prop.creation_date, doc="Date and time when this rule was created")
	modified_by = property(lambda self: self._prop.modified_by, doc="Username that last modified this rule")
	modification_date = property(lambda self: self._prop.modification_date, doc="Date and time when this rule was last modified")

	_attributes_doc = """
	Arbitrary key/value attributes stored for this rule.

	    >>> r.attributes
	    {}
	    >>> r.attributes['test'] = "Arbitrary string value"
	    >>> r.attributes
	    {'test': "Arbitrary string value"}

	@raises PLDBUnsupportedInFirmware: If attributes is accessed without firmware support.

 	@attention: Attributes requires v12.2 firmware or newer.
	"""
	attributes = property(_attr_list, doc=_attributes_doc)

	class Condition(object):
		"""
		Condition implementation

		@type op: C{int}
		@type type: C{int}
		@type fq_objects: C{list} of C{tuple} of C{int} and C{str}
		@type objects: C{list} of C{int}
		
		@cvar operators: Available operators
		@cvar conditions: Available condition types

		@raises RuntimeError: If you use objects attribute in a distributed
		        environment where mixed system id:s are present. Use the
			attribute fq_objects if multiple system id:s may occur.
		"""
		
		__slots__ = ('_pldb', '_prop')

		operators = plutil.roConstDict({
			PLDB.RULESET_CONDITION_OP_EQ: "==",
			PLDB.RULESET_CONDITION_OP_NE: "!=",
			})
		
		conditions = plutil.roConstDict({
			PLDB.RULESET_CONDITION_TYPE_NETOBJECT_CLIENT: "Client NetObject",
			PLDB.RULESET_CONDITION_TYPE_NETOBJECT_SERVER: "Server NetObject",
			PLDB.RULESET_CONDITION_TYPE_NETOBJECT_HOST: "Host NetObject",
			PLDB.RULESET_CONDITION_TYPE_NETOBJECT_LOCAL: "Local NetObject",
			PLDB.RULESET_CONDITION_TYPE_PORTOBJECT_CLIENT: "Client PortObject",
			PLDB.RULESET_CONDITION_TYPE_PORTOBJECT_SERVER: "Server PortObject",
			PLDB.RULESET_CONDITION_TYPE_SERVICEOBJECT: "ServiceObject",
			PLDB.RULESET_CONDITION_TYPE_SERVICEOBJECT_BASE: "Base ServiceObject",
			PLDB.RULESET_CONDITION_TYPE_VLANOBJECT: "VlanObject",
			PLDB.RULESET_CONDITION_TYPE_VLANOBJECT_IN: "VlanObject",
			PLDB.RULESET_CONDITION_TYPE_VLANOBJECT_OUT: "VlanObject",
			PLDB.RULESET_CONDITION_TYPE_PROTOCOLOBJECT: "ProtocolObject",
			PLDB.RULESET_CONDITION_TYPE_TIMEOBJECT: "TimeObject",
			PLDB.RULESET_CONDITION_TYPE_BGPOBJECT: "BGPObject",
			PLDB.RULESET_CONDITION_TYPE_PROPERTYOBJECT: "PropertyObject",
			PLDB.RULESET_CONDITION_TYPE_SYSTEMOBJECT: "SystemObject",
			PLDB.RULESET_CONDITION_TYPE_FLAGOBJECT: "FlagObject",
			PLDB.RULESET_CONDITION_TYPE_MPLSOBJECT: "MPLSObject",
			PLDB.RULESET_CONDITION_TYPE_MPLSOBJECT_IN: "MPLSObject",
			PLDB.RULESET_CONDITION_TYPE_MPLSOBJECT_OUT: "MPLSObject",
			PLDB.RULESET_CONDITION_TYPE_DSCPOBJECT: "DSCPObject",
			PLDB.RULESET_CONDITION_TYPE_DSCPOBJECT_IN: "DSCPObject",
			PLDB.RULESET_CONDITION_TYPE_DSCPOBJECT_OUT: "DSCPObject",
			PLDB.RULESET_CONDITION_TYPE_CHANNELOBJECT: "ChannelObject",
			PLDB.RULESET_CONDITION_TYPE_CHANNELOBJECT_IN: "ChannelObject",
			PLDB.RULESET_CONDITION_TYPE_CHANNELOBJECT_OUT: "ChannelObject",
			PLDB.RULESET_CONDITION_TYPE_TUNNELLVLOBJECT: "TunnelLevelObject",
			PLDB.RULESET_CONDITION_TYPE_TUNNELTYPEOBJECT: "TunnelTypeObject",
			})

		def __init__(self, pldb, **props):
			self._pldb = pldb
			self._prop = _Prop_condition(**props)

		def __repr__(self):
			return "<PacketLogic Condition '%s' %s %s>" % (self.conditions[self._prop.type],
								     self.operators[self._prop.op],
								     str(self._prop.objects))

		def get_objects(self):
			""" This is an undocumented method that should not be used. Use property <Condition.objects> instead."""
			# Since people using dir() to find things might have used this in code (like jek, oops :)), we keep it,
			# but issues a deprecation warning.
			import warnings
			warnings.warn("This is an undocumented method that should not be used. Use property <Condition.objects> instead.",
				      DeprecationWarning, stacklevel=2)
			return self.objects

		def _fq_objects(self):
			"""A list of (objectid, systemid)'s from the type condition."""
			return [(o, self._pldb.commitid.systemid) for o in self._prop.objects]

		systemid = plutil._sysidprop("PacketLogic SystemID")
		op = property(lambda self: self._prop.op, doc="Operator (CONDITION_OP_EQ or CONDITION_OP_NE)")
		type = property(lambda self: self._prop.type, doc="Type of condition (One of CONDITION_*)")
		objects = plutil.roCopyList('objects', "A list of object id's from the type condition.")
		fq_objects = property(_fq_objects)

class FwRule(PLRule):

	"""
	@see: L{fwrule_add <Ruleset.fwrule_add>} for more details.
	
	@type action: C{int}
	@type quick: C{bool}
	@type log: C{bool}
	@type rewriteobject: C{int}
	@type monitor_iface: C{int}
	@type monitor_label: C{str}
	@type divert_label: C{str}
	@type position: C{int}
	@type orderby: C{int}
	@type trigger: C{str}
	@type monitor_label: C{str}
	@type divert_label: C{str}

	@cvar __slots__:
	"""
	__slots__ = ('_prop', '_pldb', '_ruletype')
	
	def __init__(self, _pldb, **props):
		self._ruletype = PLDB.RULESET_RULE_TYPE_FILTER
		self._pldb = _pldb
		self._prop = _Prop_fwrule(**props)

	def update(self, **kwargs):
		"""
		Change properties of the firewall rule. The following
		properties can be changed: action, quick, log_level,
		rewriteobject, monitor_iface, trigger

			>>> rule.update(quick=True)
			>>> rule.quick
			True

			>>> rule.update(action=r.FWRULE_ACTION_ACCEPT, log_level=2)
			>>> rule.log_level
			2

		"""

		# this is the order of the arguments to _fwupdate
		keys = 'action', 'quick', 'log_level', 'rewriteobject', 'monitor_iface', 'monitor_label', 'divert_label', 'inject_data', 'trigger'

		for k in kwargs.iterkeys():
			if k not in keys:
				raise ValueError("Unknown property '%s'" % k)
			if k == "divert_iface":
				raise PLDBUnsupportedInFirmware("divert_iface is obsolete. Use divert_label instead")

		# get the argument either from kwargs or self
		args = [kwargs.get(key, getattr(self, key)) for key in keys]
		res = _fwupdate(self._pldb, self.id, *args)

		self._prop.update(**kwargs)
		self._prop.update(modification_date=res, modified_by=self._pldb.user)

	def cond_add(self, condition, operator, value):
		if condition in self._prop.conditions:
			raise ValueError("Condition already exist")

		return self._do_add(condition, operator, value,
				    self._pldb.session.fwrule_condition_set,
				    self._pldb.session.fwrule_condition_set_handle,
				    self._pldb._fwrule_invalidate_cache)

	def _cond_set(self, *args):
		return (self._pldb.session.fwrule_condition_set_handle,
			self._pldb.session.fwrule_condition_set(*args))
		
	def _cond_object_add(self, *args):
		return (self._pldb.session.fwrule_condition_object_add_handle,
			self._pldb.session.fwrule_condition_object_add(*args))

	def _cond_object_remove(self, *args):
		return (self._pldb.session.fwrule_condition_object_remove_handle,
			self._pldb.session.fwrule_condition_object_remove(*args))
	
	def cond_remove(self, condition):
		return self._do_remove(condition,
				       self._pldb.session.fwrule_condition_remove,
				       self._pldb.session.fwrule_condition_remove_handle,
				       self._pldb._fwrule_invalidate_cache)


	def enable(self):
		self._set_active(self._pldb.session.fwrule_set_active,
				 self._pldb.session.fwrule_set_active_handle,
				 True)

	def disable(self):
		self._set_active(self._pldb.session.fwrule_set_active,
				 self._pldb.session.fwrule_set_active_handle,
				 False)

	def rename(self, newname):
		_fwrename(self._pldb, self, newname)
		self._prop.name = newname

		self._pldb._invalidate_cache()
		
	def move_up(self):
		""" Move this rule one step up (to a lower position number) by swapping it with
		the neighbouring rule. Returns False if at top allready, otherwise True.

		@raise ValueError: If the rule is global.
		
		@rtype: C{bool}
		
		"""
		return self._move(-1)
	
	def move_down(self):
		""" Move this rule down step up (to a higher position number) by swapping it with
		the neighbouring rule. Returns False if at bottom allready, otherwise True.

		@raise ValueError: If the rule is global.
		@raise ValueError: If the rule nolonger exists.
		
		@rtype: C{bool}
		
		"""
		return self._move(1)

	def _move(self, direction):
		# direction is 1 for down (higher number) or -1 for up (lower number)
		
		self._pldb._fwrule_update_cache()
		# We currently can't be sure the rule list is ordered, e.g fwrule_Add will append the new rule,
		# with highest number, to the otherwise revered list with lowest number (zero) at the end.
		# Therefore we create a new list that we sort with orderby, also filtering away global rules.
		#local_rules = [(r._prop.position, r) for r in self._pldb._fwrulecache if not r._prop.is_global]
		local_rules = []
		thisrule = thisix = None
		
		for ix, r in enumerate(self._pldb._fwrulecache):
			local_rules.append((r._prop.position, r, ix))
			if r._prop.id == self._prop.id:
				# Just in case we call with a copy, use thisrule instead of self for modifications
				thisrule, thisix = r, ix

		if thisrule is None:
			raise ValueError("The rule you try to move nolonger exists in the system.")
		
		local_rules.sort()
		if direction == -1:
			local_rules.reverse()

		for (pos, otherrule, otherix) in local_rules:
			if cmp(otherrule._prop.position, self._prop.position) == direction:
				break
		else:
			# We are at the top/bottom
			return False

		tstamp = _fwswap(self._pldb, thisrule, otherrule)

		thisrule._prop.modification_date = tstamp
		thisrule._prop.modified_by = self._pldb.user
		otherrule._prop.modification_date = tstamp
		otherrule._prop.modified_by = self._pldb.user
		
		# Swap position property values, as well as rules position in the cached list
		thisrule._prop.position, otherrule._prop.position = otherrule._prop.position, thisrule._prop.position
		self._pldb._fwrulecache[thisix], self._pldb._fwrulecache[otherix] = self._pldb._fwrulecache[otherix], self._pldb._fwrulecache[thisix]
		
		# If it was a copy (usually due to something invalidating the cache), update that to.
		self._prop.modification_date = thisrule._prop.modification_date
		self._prop.modified_by = thisrule._prop.modified_by
		self._prop.position = thisrule._prop.position
		
		return True

	def _divert_iface(self):
		raise PLDBUnsupportedInFirmware("divert_iface is obsolete. Use divert_label instead")

	action = property(lambda self: self._prop.action, doc="Action when the rule matches. See L{fwrule_add <Ruleset.fwrule_add>}")
	quick = property(lambda self: self._prop.quick, doc="State of quick property\n"
			 "(True means that no more rules in the ruleset will be evaluated if this rule matches.)")
	log = property(lambda self: bool(self._prop.log_level), doc="If True, hits to this rule is logged.")
	log_level = property(lambda self: self._prop.log_level, doc="0 = Off. 1 = Brief. 2 = Verbose")
	rewriteobject = property(lambda self: self._prop.rewriteobject, doc="Zero or the id of a L{rewriteobject.RewriteObject}")
	monitor_iface = property(lambda self: self._prop.monitor_iface, doc="If non zero, packets matching this rule will also be "
				 "sent to the specified L{monitor interface <Ruleset.monitor_iface_list>} id.")
	position = property(lambda self: self._prop.position, doc="")
	orderby = property(lambda self: self._prop.orderby, doc="")

	_inject_data_doc = """The content that shall be injected for
	                      FWRULE_ACTION_INJECT rules. The content
	                      is usually a HTTP 307 redirect response.

	                      @attention: inject_data requires v12.1 firmware or newer, earlier versions always returns "".
	                      """
	inject_data = property(lambda self: self._prop.inject_data, doc=_inject_data_doc)

	trigger = property(lambda self: self._prop.trigger, doc="Empty string or name of a trigger. Since: v11")

	_monitor_label_doc = """Label to monitor packets to when monitor_iface
				is set to the index of the "Label" interface

				@attention: monitor_label requires v14.1 firmware or newer, earlier versions always returns "" """
	monitor_label = property(lambda self: self._prop.monitor_label, doc=_monitor_label_doc)

	_divert_label_doc = """Label to divert packets to when
	                       using FWRULE_ACTION_DIVERT.

	                       @attention: divert_label requires v14.1 firmware or newer, earlier versions always returns "".
	                       """
	divert_label = property(lambda self: self._prop.divert_label, doc=_divert_label_doc)

	_divert_iface_doc = """Interface to divert packets to when
	                       using FWRULE_ACTION_DIVERT. See
	                       L{divert_iface_list
	                       <Ruleset.divert_iface_list>}.

	                       @attention: divert_iface is only used in firmwares 14.0 or older. In 14.1 and newer use divert_label.
	                       """
	divert_iface = property(_divert_iface, doc=_divert_iface_doc)

	_monitor_label_doc = """Label to monitor packets to when monitor_iface
			     is set to the index of the "Label" interface.

			     @attention: monitor_label requires v14.1 firmware or newer, earlier versions always returns ""

                             """
	monitor_label = property(lambda self: "", doc=_monitor_label_doc)

	_divert_label_doc = """Label to divert packets to when
	                       using FWRULE_ACTION_DIVERT.

	                       @attention: divert_label requires v14.1 firmware or newer, earlier versions always returns "".
	                       """
	divert_label = property(lambda self: "", doc=_divert_label_doc)


class ShapingRule(PLRule):

	"""
	@note: The flags are currently available as properties
	on the ShapingRule object directly, as e.g rule.exclusive,
	but in future versions that might change, and the use of
	e.g rule.flags.exclusive is recommended.

	@type priority: C{int}
	@type objects: read-only C{list}
	@type flags: L{ShapingRuleFlags} instance
	
	@deprecated: The max_connections argument is depraceted since version 11.x
	The property will allways return zero, and issue a deprecation warning if accessed.
	It is replaced by the max_connections property on L{shapingobjects <shapingobject.ShapingObject>}.
	
	@cvar __slots__: For internal use only.
	"""
	
	__slots__ = ('_prop', '_pldb', '_ruletype')
	
	def __init__(self, _pldb, **props):
		self._ruletype = PLDB.RULESET_RULE_TYPE_SHAPING
		self._pldb = _pldb
		self._prop = _Prop_shapingrule(**props)

	def cond_add(self, condition, operator, value):
		if condition in self._prop.conditions:
			raise ValueError("Condition already exist")

		return self._do_add(condition, operator, value,
				    self._pldb.session.shapingrule_condition_set,
				    self._pldb.session.shapingrule_condition_set_handle,
				    self._pldb._shapingrule_invalidate_cache)

	def _cond_set(self, *args):
		return (self._pldb.session.shapingrule_condition_set_handle,
			self._pldb.session.shapingrule_condition_set(*args))

	def _cond_object_add(self, *args):
		return (self._pldb.session.shapingrule_condition_object_add_handle,
			self._pldb.session.shapingrule_condition_object_add(*args))

	def _cond_object_remove(self, *args):
		return (self._pldb.session.shapingrule_condition_object_remove_handle,
			self._pldb.session.shapingrule_condition_object_remove(*args))
	
	def cond_remove(self, condition):
		return self._do_remove(condition,
				       self._pldb.session.shapingrule_condition_remove,
				       self._pldb.session.shapingrule_condition_remove_handle,
				       self._pldb._shapingrule_invalidate_cache)

	def set_objects(self, objects):
		"""
		Sets list of shapingobjects in this rule

		@type objects: C{list} of C{int}
		@param objects: A list of L{ShapingObject IDs <shapingobject.ShapingObject.id>}

		"""

		cid = self._pldb.session.shapingrule_set_objects(self._prop.id, objects)
		tstamp = plutil._do(self._pldb, cid, plutil._AddCB(self._pldb, self._pldb.session.shapingrule_set_objects_handle))
		self._prop.objects = plutil.RoList(objects)
		self._prop.modification_date = tstamp
		self._prop.modified_by = self._pldb.user

	def enable(self):
		self._set_active(self._pldb.session.shapingrule_set_active,
				 self._pldb.session.shapingrule_set_active_handle,
				 True)

	def disable(self):
		self._set_active(self._pldb.session.shapingrule_set_active,
				 self._pldb.session.shapingrule_set_active_handle,
				 False)

	def set_flags(self, *reset, **flags):
		"""
		Set or reset one or more of the flags of the shaping rule.
		You may also reset all flags with the single argument None, or clone an existing flags object.
		You may combine them.

		You read the flags with e.g 'excl = myRule.flags.exclusive', but in the current version you
		can still access them the old way directly with e.g 'excl = myRule.exclusive'.

		If only keyword argument(s) are used, only the specified flags will be affected.
		Use a trueish value to set the flag, or falseish value to reset it.

			>>> shrule.set_flags(exclusive=1, stats=False)
			>>> shrule.flags
			ShapingRuleFlags(exclusive=True)
			
			>>> shrule.flags.exclusive
			True
			
			>>> shrule.set_flags(None)
			>>> shrule.flags
			ShapingRuleFlags(None)

			>>> shrule.set_flags(another_shrule)
			>>> shrule.flags
			ShapingRuleFlags(stats=True)
			
			>>> shrule.set_flags(None, exclusive=True)
			>>> shrule.flags
			ShapingRuleFlags(exclusive=True)

		@keyword exclusive: This is an exclusive rule.
		@keyword stats: Enable or disable statistics for traffic matching the rule
		@keyword stats_netobjects: Enable or disable separation of statistics into netobjects
		@keyword stats_services: Enable or disable separation of statistics into services
		@keyword stats_local_vhosts: Enable or disable separation of statistics into local vhosts
		@keyword stats_remote_vhosts: Enable or disable separation of statistics into remote vhosts

		"""
		
		if len(reset) == 1 and isinstance(reset[0], self.__class__):
			self.flags._set_flags(reset[0].flags, **flags)

		else:
			self.flags._set_flags(*reset, **flags)

		c = self._pldb.session.shapingrule_update(self._prop.id, self._prop.priority, **dict(self._prop.flags.items()))
		plutil._do(self._pldb, c, plutil._AddCB(self._pldb, self._pldb.session.shapingrule_update_handle))

	### The below method will soon be deprecated in favour for set_flags FIXME ###
	def set_stats(self, stats=None, stats_netobjects=None, stats_services=None, stats_local_vhosts=None, stats_remote_vhosts=None):
		"""
		Enables or disables different types of statistics
		collection for the rule. Specify only the types you
		want to change

		     >>> s=r.shapingrule_find("stats")
		     >>> s.set_stats(stats=True, stats_services=True)

		     >>> #disable services stats, enable netobject stats
		     >>> s=r.shapingrule_find("stats")
		     >>> s.set_stats(stats=True, stats_services=False, stats_netobjects=True)

		@type stats: C{bool}
		@param stats: Enable or disable statistics for traffic matching the rule

		@type stats_netobjects: C{bool}
		@param stats_netobjects: Enable or disable separation of statistics into netobjects

		@type stats_services: C{bool}
		@param stats_services: Enable or disable separation of statistics into services

		@type stats_local_vhosts: C{bool}
		@param stats_local_vhosts: Enable or disable separation of statistics into local vhosts

		@type stats_remote_vhosts: C{bool}
		@param stats_remote_vhosts: Enable or disable separation of statistics into remote vhosts

		"""

		kwargs = {}
		if stats is not None:
			kwargs['stats'] = stats
		if stats_netobjects is not None:
			kwargs['stats_netobjects'] = stats_netobjects
		if stats_services is not None:
			kwargs['stats_services'] = stats_services
		if stats_local_vhosts is not None:
			kwargs['stats_local_vhosts'] = stats_local_vhosts
		if stats_remote_vhosts is not None:
			kwargs['stats_remote_vhosts'] = stats_remote_vhosts
			
		self.set_flags(**kwargs)

	def rename(self, newname):
		_shapingrename(self._pldb, self, newname)
		self._prop.name = newname

		self._pldb._invalidate_cache()

	def _max_connections(self):
		""" Property 'max_connections' on shapingrules are deprecated. Used on shapingobjects instead."""
		import warnings
		warnings.warn("Property 'max_connections' on shapingrules are deprecated. Used on shapingobjects instead.",
			      DeprecationWarning, stacklevel=2)
		return 0

	max_connections = property(_max_connections)
	
	priority = property(lambda self: self._prop.priority, doc="The priority of packets matching this rule.")
	objects = property(lambda self: self._prop.objects, doc="List of L{shaping objects <shapingobject.ShapingObject>} in this rule.\n"
				"Changed with L{set_objects() <plrule.ShapingRule.set_objects>}.")
	flags = property(lambda self: self._prop.flags, doc="Collection of flags on this rule.\n"
			 "Also see L{set_flags() <plrule.ShapingRule.set_flags>}.")

	### below should soon be deprecated in favour for flags FIXME ###
	
	exclusive = property(lambda self: self._prop.flags.exclusive,
			     doc="This is an exclusive rule. You should use flags.exclusive")
	stats = property(lambda self: self._prop.flags.stats,
			 doc="Enable statistics for traffic matching the rule. You should use flags.stats")
	stats_netobjects = property(lambda self: self._prop.flags.stats_netobjects,
				    doc="Enable separation of statistics into netobjects. You should use flags.stats_netobjects")
	stats_services = property(lambda self: self._prop.flags.stats_services,
				  doc="Enable separation of statistics into services. You should use flags.stats_services")
	stats_local_vhosts = property(lambda self: self._prop.flags.stats_local_vhosts,
				      doc="Enable separation of statistics into local vhosts. You should use flags.stats_local_vhosts")
	stats_remote_vhosts = property(lambda self: self._prop.flags.stats_remote_vhosts,
				       doc="Enable separation of statistics into remote vhosts. You should use flags.stats_remote_vhosts")
	### above should soon be deprecated in favour for flags ###
	
_Prop_monitorIface = plutil.Prop('id', 'description', 'iface', 'media')
class MonitorIface(object):
	"""
	Object representing a monitor interfaces.

	Returned in a list by L{Ruleset.monitor_iface_list() <Ruleset.monitor_iface_list>}.
	
	@type id: C{int}
	@type description: C{str}
	@type iface: C{str}
	@type media: C{int}
	
	"""
	__slots__ = ('_prop')

	def __init__(self, **props):
		props.setdefault('description', '')
		self._prop = _Prop_monitorIface(**props)
		
	def __repr__(self):
		return "<PacketLogic Monitor Interface '%s'>" % (self.description)

	id = property(lambda self: self._prop.id, doc="Id number, used as argument to L{fwrule_add <Ruleset.fwrule_add>}.")
	description = property(lambda self: self._prop.description, doc="Description of the monitor interface")
	iface = property(lambda self: self._prop.iface, doc="Name of the internal network interface")
	media = property(lambda self: self._prop.media, doc="Network media type")

class StatisticsRule(PLRule):
	"""
	@type objectid: read-only C{list}
	@type flags: L{StatisticsRuleFlags} instance
	
	@cvar __slots__: For internal use only.
	"""
	
	__slots__ = ('_prop', '_pldb', '_ruletype')
	
	def __init__(self, _pldb, **props):
		self._ruletype = PLDB.RULESET_RULE_TYPE_STATISTICS
		self._pldb = _pldb
		self._prop = _Prop_statisticsrule(**props)

	def cond_add(self, condition, operator, value):
		if condition in self._prop.conditions:
			raise ValueError("Condition already exist")

		return self._do_add(condition, operator, value,
				    self._pldb.session.statisticsrule_condition_set,
				    self._pldb.session.statisticsrule_condition_set_handle,
				    self._pldb._statisticsrule_invalidate_cache)

	def _cond_set(self, *args):
		return (self._pldb.session.statisticsrule_condition_set_handle,
			self._pldb.session.statisticsrule_condition_set(*args))

	def _cond_object_add(self, *args):
		return (self._pldb.session.statisticsrule_condition_object_add_handle,
			self._pldb.session.statisticsrule_condition_object_add(*args))

	def _cond_object_remove(self, *args):
		return (self._pldb.session.statisticsrule_condition_object_remove_handle,
			self._pldb.session.statisticsrule_condition_object_remove(*args))
	
	def cond_remove(self, condition):
		return self._do_remove(condition,
				       self._pldb.session.statisticsrule_condition_remove,
				       self._pldb.session.statisticsrule_condition_remove_handle,
				       self._pldb._statisticsrule_invalidate_cache)

	def enable(self):
		self._set_active(self._pldb.session.statisticsrule_set_active,
				 self._pldb.session.statisticsrule_set_active_handle,
				 True)

	def disable(self):
		self._set_active(self._pldb.session.statisticsrule_set_active,
				 self._pldb.session.statisticsrule_set_active_handle,
				 False)

	def set_flags(self, *reset, **flags):
		"""
		Set or reset one or more of the flags of the statistics rule.
		"""
		if len(reset) == 1 and isinstance(reset[0], self.__class__):
			self.flags._set_flags(reset[0].flags, **flags)

		else:
			self.flags._set_flags(*reset, **flags)

		c = self._pldb.session.statisticsrule_update(self._prop.id, **dict(self._prop.flags.items()))
		plutil._do(self._pldb, c, plutil._AddCB(self._pldb, self._pldb.session.statisticsrule_update_handle))

	def set_object(self, objectid):
		"""
		Sets L{statisticsobject <statisticsobject.StatisticsObject>} used by this rule
		"""
		self.set_objects ([objectid])

	def set_objects(self, objects):
		"""
		Sets list of statisticsobjects in this rule

		@type objects: C{list} of C{int}
		@param objects: A list of L{StatisticsObject IDs <statisticsobject.StatisticsObject.id>}

		@raises PLDBUnsupportedInFirmware: If this function is called without firmware support.

		@attention: This function requires firmware v14.1 or newer
		"""

		cid = self._pldb.session.statisticsrule_set_objects(self._prop.id, objects)
		tstamp = plutil._do(self._pldb, cid, plutil._AddCB(self._pldb, self._pldb.session.statisticsrule_set_objects_handle))
		self._prop.objects = plutil.RoList(objects)
		self._prop.modification_date = tstamp
		self._prop.modified_by = self._pldb.user

	def rename(self, newname):
		_statisticsrename(self._pldb, self, newname)
		self._prop.name = newname
		self._pldb._invalidate_cache()

	objectid = property(lambda self: self._prop.objectid, doc="Id number of L{statistics object <statisticsobject.StatisticsObject>} used in this rule.\n")
	objects = property(lambda self: self._prop.objects, doc="List of L{statistics objects <statisticsobject.StatisticsObject>} in this rule.\n"
				"Changed with L{set_objects() <plrule.StatisticsRule.set_objects>}.")

	flags = property(lambda self: self._prop.flags, doc="Collection of flags on this rule.\n"
			 "Also see L{set_flags() <plrule.StatisticsRule.set_flags>}.")
